/*
   Pre-process database routines, using a pre-processor specific memory mapped
   database.
 */

/* Copyright, compile requirements.

   (c) Keith Owens <kaos@ocs.com.au> January 2002.
   Released under GPL version 2, (you know the rest).
 */

#define NEED_PPDB_INTERNALS
#include "pp_makefile.h"
#include "pp_db.h"
#include <limits.h>
#include <time.h>
#include <signal.h>

extern const char *program;
extern int norun;
extern int signalled;

static int pagesizem1;		/* Page size - 1 */

/*
 *==============================================================================
 *
 * A word of warning about database records.  Many functions return pointers
 * into the database pages.  These pointers are always valid for fixed sized
 * records and for fixed and variable sized keys.  Only variable sized records
 * can move after they are initially created.  A variable size record will only
 * move if it is expanded, and only the affected record will move.
 *
 * If in doubt, take a local copy of the data, but then any changes must be
 * manually put back into the database.
 *
 * It is often a mistake to pass a pointer to a database record into a function
 * unless (a) the pointer is a local copy or (b) you know that the function does
 * not expand records of the same type as the pointer refers to.
 *
 *==============================================================================
 */

static const DB_TOKEN db_null_token = DB_TOKEN_NULL;
const DB_TOKEN key_base = DB_KEY_BASE_LIST;

DB_REC_BASE_LIST *rec_base;

/*============================================================================*/

/* Strip trailing '/' and return strlen */

static int
strip_trailing(char *name, int l)
{
	assert(name[0] == '/');
	while (l-- && name[l] == '/') {
	};
	if (l < 0) {
		l = 0;
	}
	name[++l] = '\0';
	return (l);
}

/* Assign the next record number as a key of the desired type. */

static void
assign_db_key(PPDB * db, DB_TYPE type, DB_TOKEN * token)
{
	if (db->header->next_recno >= ppdb_fixed_key_max) {
		fprintf(stderr,
			"%s: assign_db_key - all records used, should never occur\n",
			program);
		abort();
	}
	token->type = type;
	token->recno = (db->header->next_recno)++;
}

/* Unlink a file and clear its time in the database.  All unlinks are done
 * against the object tree.  Gets rid of both source and target files with
 * the specified name.
 */

void
unlink_db_files(PPDB * db, const char *relname)
{
	DB_TOKEN token;
	int ret, deleted = 0, l = l_objtree + strlen(relname + 1) + 1;
	char name[l];
	memcpy(name, objtree, l_objtree);
	strcpy(name + l_objtree, relname + 1);
	VERBOSE(2, "    unlink_db_files %s\n", relname + 1);
	token =
	    lookup_db_byname(db, relname, DB_RECTYPE_SOURCE,
			     objtree_treename_index);
	if (!DB_IS_NULL_TOKEN(token)) {
		DB_REC_SOURCE *s = read_db_source(db, token);
		if (s->mtime && !norun) {
			ret = unlink(name);
			assert(ret == 0);
		}
		deleted = 1;
		s->mtime = 0;
	}
	token = lookup_db_target(db, relname);
	if (!DB_IS_NULL_TOKEN(token)) {
		DB_REC_TARGET *t = read_db_target(db, token);
		if (t->mtime && !norun && !deleted) {
			ret = unlink(name);
			assert(ret == 0);
		}
		deleted = 1;
		t->mtime = 0;
		t->flags.uptodate = 0;
		t->flags.symlink = 0;
	}
}

/*============================================================================*/

/* Read a chunk of data safely */

static int
safe_read(int fd, off_t offset, void *data, size_t size, const char *text)
{
	ssize_t i;
	size_t l = size;
	if (lseek(fd, offset, SEEK_SET) != offset) {
		fprintf(stderr,
			"%s: safe_read for %s seek to %lld failed: %m\n",
			program, text, (long long) offset);
		abort();
	}
	while (l) {
		i = read(fd, data, l);
		if (i < 0) {
			if (errno == EINTR) {
				continue;
			}
			fprintf(stderr, "%s: safe_read for %s failed: %m\n",
				program, text);
			abort();
		}
		if (i == 0) {
			break;
		}
		l -= i;
		(char *) data += i;
	}
	return (l ? 1 : 0);
}

/* Write a chunk of data safely */

static void
safe_write(int fd, off_t offset, void *data, size_t size, const char *text)
{
	ssize_t i;
	size_t l = size;
	if (lseek(fd, offset, SEEK_SET) != offset) {
		fprintf(stderr,
			"%s: safe_write for %s seek to %lld failed: %m\n",
			program, text, (long long) offset);
		abort();
	}
	while (l) {
		i = write(fd, data, l);
		if (i < 0) {
			if (errno == EINTR) {
				continue;
			}
			fprintf(stderr, "%s: safe_write for %s failed: %m\n",
				program, text);
			abort();
		}
		if (i == 0) {
			break;
		}
		l -= i;
		(char *) data += i;
	}
	if (l) {
		fprintf(stderr,
			"%s: safe_write for %s is incomplete, %lld bytes left\n",
			program, text, (long long) l);
		abort();
	}
}

/*============================================================================*/

/* Calculate a hash for a string */

static unsigned int
hash(const char *name)
{
	unsigned int h = 0, a = 255;
	for (; *name; ++name) {
		h = (a * h + *name) % ppdb_hashsize;
	}
	return h;
}

/* Open the database for action */

#define array_size(x) (sizeof(x)/sizeof((x)[0]))
#define adjust_pointer(x,y) if (x) (char *)(x) += y

static void
ppdb_open(PPDB * db)
{
	struct stat st;
	int i, mapsize, dirty;
	void *map;
	long long adjust;
	if (fstat(db->fd, &st)) {
		fprintf(stderr, "%s: fstat(%s) failed: %m\n", program,
			ppdb_name);
		abort();
	}
	if (st.st_size == 0) {
		fprintf(stderr, "%s: %s is empty: %m\n", program, ppdb_name);
		abort();
	}
	mapsize = (st.st_size + pagesizem1) & ~pagesizem1;
	if (mapsize != st.st_size) {
		int l = mapsize - st.st_size;
		void *add = k_malloc(l);
		memset(add, 0, l);
		safe_write(db->fd, st.st_size, add, l, "ppdb_open");
		free(add);
	}
	map =
	    mmap(NULL, mapsize, PROT_READ | PROT_WRITE, MAP_SHARED, db->fd, 0);
	if ((long) map == -1) {
		fprintf(stderr, "%s: mmap(%s) failed: %m\n", program,
			ppdb_name);
		abort();
	}
	if ((unsigned long) map & pagesizem1) {
		fprintf(stderr, "%s: mmap(%s) not aligned: %m\n", program,
			ppdb_name);
		abort();
	}
	db->header = map;
	dirty = db->header->flags.dirty;
	db->header->flags.dirty = 1;
	db->header->size = mapsize;
	adjust = (char *) (db->header) - (char *) (db->header->self);
	for (i = 0; i < array_size(db->header->key_fixed); ++i) {
		adjust_pointer(db->header->key_fixed[i], adjust);
	}
	adjust_pointer(db->header->source_slab, adjust);
	adjust_pointer(db->header->target_slab, adjust);
	adjust_pointer(db->header->directory_slab, adjust);
	adjust_pointer(db->header->free_start, adjust);
	adjust_pointer(db->header->self, adjust);
	db->header->free_left =
	    mapsize - ppdb_offset(db, db->header->free_start);
	db->header->flags.dirty = dirty;
}

/* Close the database, the file is left open */

static void
ppdb_close(PPDB * db)
{
	/* Dummy write to update timestamp on database */
	safe_write(db->fd, 0, db->header, 1, "ppdb_close");
	if (msync(db->header, db->header->size, MS_ASYNC)) {
		fprintf(stderr, "%s: ppdb_close msync failed: %m\n",
			program);
		abort();
	}
	if (munmap(db->header, db->header->size)) {
		fprintf(stderr, "%s: munmap(%s) failed: %m\n", program,
			ppdb_name);
		abort();
	}
	db->header = NULL;
}

/* Claim a chunk of free space */

static void *
ppdb_free_space(PPDB * db, int size)
{
	void *p = db->header->free_start, *add;
	struct stat st;
	int l, oldsize;
	void *map, *oldmap;
	db->header->free_start += size;
	db->header->free_left -= size;
	if (db->header->free_left >= 0) {
		return (p);
	}

	/* Need to extend the database and remap it */
	l = (db->header->size + ppdb_extend + pagesizem1) & ~pagesizem1;
	add = k_malloc(l);
	memset(add, 0, l);
	safe_write(db->fd, db->header->size, add, l, "ppdb_free_space");
	free(add);
	db->header->free_left += l;
	oldsize = db->header->size;
	if (msync(db->header, db->header->size, MS_SYNC)) {
		fprintf(stderr, "%s: ppdb_free_space msync failed: %m\n",
			program);
		abort();
	}
	oldmap = db->header;
	ppdb_close(db);
	if (munmap(oldmap, oldsize)) {
		fprintf(stderr, "%s: munmap(%s) failed: %m\n", program,
			ppdb_name);
		abort();
	}
	if (fstat(db->fd, &st)) {
		fprintf(stderr, "%s: fstat(%s) failed: %m\n", program,
			ppdb_name);
		abort();
	}
	map = mmap(oldmap, st.st_size, PROT_READ | PROT_WRITE,
		   MAP_SHARED | MAP_FIXED, db->fd, 0);
	if ((long) map == -1) {
		fprintf(stderr, "%s: mmap(%s) failed: %m\n", program,
			ppdb_name);
		abort();
	}
	if (map != oldmap) {
		fprintf(stderr,
			"%s: mmap(%s) worked but address is wrong 0x%p 0x%p\n",
			program, ppdb_name, oldmap, map);
		abort();
	}
	db->header = map;
	db->header->size = st.st_size;
	return (p);
}

/*============================================================================*/

/* Put a variable value on the free chain for its size */

static void
put_free_val(PPDB * db, struct ppdb_value_variable *v)
{
	int i = v->order;
	v->vlen = db->header->variable_free[i];
	db->header->variable_free[i] = ppdb_offset(db, v);
}

/* Get the next free slot to hold a value of the desired size.  Caller must set
 * vlen, kvoff.
 */

static struct ppdb_value_variable *
get_free_val(PPDB * db, int size)
{
	struct ppdb_value_variable *v;
	int i = ppdb_order(size);
	assert(i < array_size(db->header->variable_free));
	if (db->header->variable_free[i]) {
		v = ppdb_pointer(db, db->header->variable_free[i]);
		db->header->variable_free[i] = v->vlen;
	} else {
		v = ppdb_free_space(db, 1 << (i + ppdb_variable_shift_min));
		v->order = i;
	}
	return (v);
}

/* Get the next free slot to hold a key of the desired size.  Caller must set
 * klen, collision, voff.
 */

static struct ppdb_key_variable *
get_free_key(PPDB * db, int size)
{
	struct ppdb_key_variable *k;
	int i = ppdb_order(size);
	assert(i < array_size(db->header->variable_free));
	if (db->header->variable_free[i]) {
		k = ppdb_pointer(db, db->header->variable_free[i]);
		db->header->variable_free[i] = k->klen;
	} else {
		k = ppdb_free_space(db, 1 << (i + ppdb_variable_shift_min));
		k->order = i;
	}
	return (k);
}

/*============================================================================*/

/* Return the next key in the database.  All the fixed keys first, in ascending
 * order then all the variable keys, in hash chain order.
 */

static int
ppdb_next_fixed(PPDB * db, kvpair * kv)
{
	int i, j;
	struct ppdb_key_fixed *f;
	struct ppdb_value_variable *vv;
	++(db->key_no);
	i = db->key_no >> ppdb_slab_shift;
	j = db->key_no & (ppdb_slab_num - 1);
	if (i > ppdb_fixed_key_vector_size || !(db->header->key_fixed[i])
	    || !((f = db->header->key_fixed[i] + j)->voff)) {
		db->key_type = 2;	/* variable next */
		db->key_no = -1;	/* hash bucket number */
		return (0);
	}
	kv->key.dptr = &(f->token);
	kv->key.dsize = sizeof (f->token);
	kv->val.dptr = ppdb_pointer(db, f->voff);
	switch (f->token.type) {
	case DB_RECTYPE_SOURCE:
		kv->val.dsize = sizeof (DB_REC_SOURCE);
		break;
	case DB_RECTYPE_TARGET:
		kv->val.dsize = sizeof (DB_REC_TARGET);
		break;
	case DB_RECTYPE_DIRECTORY:
		kv->val.dsize = sizeof (DB_REC_DIRECTORY);
		break;
	default:
		vv = (struct ppdb_value_variable *) (kv->val.dptr);
		kv->val.dsize = vv->vlen;
		((char *) kv->val.dptr) += offsetof(typeof(*vv), value);
	}
	return (1);
}

static int
ppdb_next_variable(PPDB * db, kvpair * kv)
{
	struct ppdb_key_variable *v;
	struct ppdb_value_variable *vv;
	void *p = NULL;
	if (db->var_key && db->var_key->collision) {
		p = ppdb_pointer(db, db->var_key->collision);
	} else {
		while (++(db->key_no) < array_size(db->header->hash)
		       && db->header->hash[db->key_no] == 0) {
			/* Find next hash bucket in use */
		};
		if (db->key_no < array_size(db->header->hash)) {
			p = ppdb_pointer(db, db->header->hash[db->key_no]);
		}
	}
	if (!p) {
		db->key_type = 0;	/* end of keys */
		return (0);
	}
	v = p;
	kv->key.dptr = &(v->value);
	kv->key.dsize = v->klen;
	kv->val.dptr = ppdb_pointer(db, v->voff);
	vv = (struct ppdb_value_variable *) (kv->val.dptr);
	kv->val.dsize = vv->vlen;
	((char *) kv->val.dptr) += offsetof(typeof(*vv), value);
	return (1);
}

kvpair
ppdb_next(PPDB * db)
{
	kvpair kv;
	if (db->key_type == 1 && ppdb_next_fixed(db, &kv)) {
		return (kv);
	}
	if (db->key_type == 2 && ppdb_next_variable(db, &kv)) {
		return (kv);
	}
	memset(&kv, 0, sizeof (kv));
	return (kv);
}

/* Return the first key in the database */

kvpair
ppdb_first(PPDB * db)
{
	db->key_type = 1;
	db->key_no = 0;
	db->var_key = NULL;
	return (ppdb_next(db));
}

/*============================================================================*/

/* Fetch a database record given the key */

static inline datum
ppdb_fetch_fixed(PPDB * db, datum key, DB_TOKEN * t)
{
	struct ppdb_key_fixed *f;
	struct ppdb_value_variable *vv;
	int i, j;
	datum val;
	i = t->recno >> ppdb_slab_shift;
	j = t->recno & (ppdb_slab_num - 1);
	if (i > ppdb_fixed_key_vector_size || !(db->header->key_fixed[i])
	    || !((f = db->header->key_fixed[i] + j)->voff)
	    || memcmp(&(f->token), t, sizeof (*t))) {
		fprintf(stderr,
			"%s: Attempt to fetch invalid key %c(0x%x)-%d\n",
			program, t->type, t->type, t->recno);
		abort();
	}
	val.dptr = ppdb_pointer(db, f->voff);
	switch (f->token.type) {
	case DB_RECTYPE_SOURCE:
		val.dsize = sizeof (DB_REC_SOURCE);
		break;
	case DB_RECTYPE_TARGET:
		val.dsize = sizeof (DB_REC_TARGET);
		break;
	case DB_RECTYPE_DIRECTORY:
		val.dsize = sizeof (DB_REC_DIRECTORY);
		break;
	default:
		vv = (struct ppdb_value_variable *) (val.dptr);
		val.dsize = vv->vlen;
		((char *) val.dptr) += offsetof(typeof(*vv), value);
	}
	return (val);
}

static inline datum
ppdb_fetch_variable(PPDB * db, datum key)
{
	datum val;
	struct ppdb_key_variable *k;
	struct ppdb_value_variable *vv;
	int i;
	memset(&val, 0, sizeof (val));
	i = hash(key.dptr);
	if (!db->header->hash[i]) {
		return (val);
	}
	k = ppdb_pointer(db, db->header->hash[i]);
	while (k) {
		if (k->klen == key.dsize
		    && memcmp(k->value, key.dptr, key.dsize) == 0) {
			break;
		}
		if (k->collision) {
			k = ppdb_pointer(db, k->collision);
		} else {
			k = NULL;
		}
	}
	if (k) {
		val.dptr = ppdb_pointer(db, k->voff);
		vv = (struct ppdb_value_variable *) (val.dptr);
		val.dsize = vv->vlen;
		((char *) val.dptr) += offsetof(typeof(*vv), value);
	}
	return (val);
}

datum
ppdb_fetch(PPDB * db, datum key)
{
	DB_TOKEN *t = key.dptr;
	if (ppdb_fixed_key(t->type)) {
		return (ppdb_fetch_fixed(db, key, t));
	} else {
		return (ppdb_fetch_variable(db, key));
	}
}

/*============================================================================*/

/* Store functions */

/* Store a fixed sized value in the corresponding slab */

static inline void
ppdb_store_fixed_val(PPDB * db, struct ppdb_key_fixed *f, datum val,
		     char **slab, int *slab_free)
{
	char *data;
	if (f->voff) {
		data = ppdb_pointer(db, f->voff);
	} else {
		if (*slab_free == 0) {
			*slab = ppdb_free_space(db, val.dsize * ppdb_slab_num);
			*slab_free = ppdb_slab_num;
		}
		data = *slab + (ppdb_slab_num - *slab_free) * val.dsize;
		f->voff = ppdb_offset(db, data);
		--*slab_free;
	}
	memcpy(data, val.dptr, val.dsize);
}

/* Store a fixed size key, the value may be fixed or variable size */

static inline void
ppdb_store_fixed_key(PPDB * db, datum key, datum val, unsigned int flags,
		     DB_TOKEN * t)
{
	struct ppdb_key_fixed *f = NULL;
	struct ppdb_value_variable *vv, *vv_old;
	int i, j, l = val.dsize + offsetof(typeof(*vv), value);
	i = t->recno >> ppdb_slab_shift;
	j = t->recno & (ppdb_slab_num - 1);
	if (i > ppdb_fixed_key_vector_size) {
		fprintf(stderr,
			"%s: Attempt to store invalid key %c(0x%x)-%d\n",
			program, t->type, t->type, t->recno);
		abort();
	}
	if (db->header->key_fixed[i] == NULL) {
		db->header->key_fixed[i] =
		    ppdb_free_space(db, ppdb_slab_num * sizeof (*f));
	}
	f = db->header->key_fixed[i] + j;
	if (flags & PPDB_INSERT) {
		if (f->voff) {
			fprintf(stderr,
				"%s: Attempt to insert existing key %c(0x%x)-%d\n",
				program, t->type, t->type, t->recno);
			abort();
		}
		memcpy(&(f->token), t, sizeof (*t));
	} else if (flags & PPDB_REPLACE) {
		if (memcmp(&(f->token), t, sizeof (*t))) {
			fprintf(stderr,
				"%s: Attempt to replace non-existent key %c(0x%x)-%d\n",
				program, t->type, t->type, t->recno);
			abort();
		}
	}
	switch (f->token.type) {
	case DB_RECTYPE_SOURCE:
		assert(val.dsize == sizeof (DB_REC_SOURCE));
		ppdb_store_fixed_val(db, f, val,
				     (char **) &(db->header->source_slab),
				     &(db->header->source_slab_free));
		break;
	case DB_RECTYPE_TARGET:
		assert(val.dsize == sizeof (DB_REC_TARGET));
		ppdb_store_fixed_val(db, f, val,
				     (char **) &(db->header->target_slab),
				     &(db->header->target_slab_free));
		break;
	case DB_RECTYPE_DIRECTORY:
		assert(val.dsize == sizeof (DB_REC_DIRECTORY));
		ppdb_store_fixed_val(db, f, val,
				     (char **) &(db->header->directory_slab),
				     &(db->header->directory_slab_free));
		break;
	default:
		/* Variable sized value */
		if (f->voff) {
			vv_old = ppdb_pointer(db, f->voff);
		} else {
			vv_old = NULL;
		}
		if (vv_old
		    && l <= (1 << (vv_old->order + ppdb_variable_shift_min))) {
			vv = vv_old;	/* reuse existing slot */
		} else {
			vv = get_free_val(db, l);
		}
		vv->vlen = val.dsize;
		memcpy(&(vv->value), val.dptr, val.dsize);
		if (vv_old && vv != vv_old) {
			put_free_val(db, vv_old);
		}
		f->voff = ppdb_offset(db, vv);
		vv->kvoff = ppdb_offset(db, &(f->voff));
	}
	return;
}

/* Store a variable size key, the value is always variable size */

static inline void
ppdb_store_variable_key(PPDB * db, datum key, datum val, unsigned int flags)
{
	struct ppdb_key_variable *k = NULL;
	struct ppdb_value_variable *vv, *vv_old;
	int i, l = val.dsize + offsetof(typeof(*vv), value);
	i = hash(key.dptr);
	if (db->header->hash[i]) {
		k = ppdb_pointer(db, db->header->hash[i]);
	}
	while (k) {
		if (k->klen == key.dsize
		    && memcmp(k->value, key.dptr, key.dsize) == 0) {
			break;
		}
		if (k->collision) {
			k = ppdb_pointer(db, k->collision);
		} else {
			k = NULL;
		}
	}
	if (flags & PPDB_INSERT) {
		if (k) {
			fprintf(stderr,
				"%s: Attempt to insert existing key '%s'\n",
				program, (char *) (key.dptr));
			abort();
		}
		k = get_free_key(db, key.dsize + offsetof(typeof(*k), value));
		k->klen = key.dsize;
		k->collision = db->header->hash[i];
		k->voff = 0;
		db->header->hash[i] = ppdb_offset(db, k);
		memcpy(&(k->value), key.dptr, key.dsize);
	} else if (flags & PPDB_REPLACE) {
		if (!k) {
			fprintf(stderr,
				"%s: Attempt to replace non-existent key '%s'\n",
				program, (char *) (key.dptr));
			abort();
		}
	}
	if (k->voff) {
		vv_old = ppdb_pointer(db, k->voff);
	} else {
		vv_old = NULL;
	}
	if (vv_old && l <= (1 << (vv_old->order + ppdb_variable_shift_min))) {
		vv = vv_old;	/* reuse existing slot */
	} else {
		vv = get_free_val(db, l);
	}
	vv->vlen = val.dsize;
	memcpy(&(vv->value), val.dptr, val.dsize);
	if (vv_old && vv != vv_old) {
		put_free_val(db, vv_old);
	}
	k->voff = ppdb_offset(db, vv);
	vv->kvoff = ppdb_offset(db, &(k->voff));
	return;
}

/* Store a database record, given key and value */

void
ppdb_store(PPDB * db, datum key, datum val, unsigned int flags)
{
	DB_TOKEN *t = key.dptr;
	if (!(flags & (PPDB_INSERT | PPDB_REPLACE))) {
		fprintf(stderr,
			"%s: ppdb_store invalid flags for %c(0x%x)-%d\n",
			program, t->type, t->type, t->recno);
		abort();
	}
	if (ppdb_fixed_key(t->type)) {
		ppdb_store_fixed_key(db, key, val, flags, t);
	} else {
		ppdb_store_variable_key(db, key, val, flags);
	}
}

/*============================================================================*/

/* Expand a record, in place if possible.  Return a pointer to the record,
 * which may have moved.
 */

static void *
ppdb_expand_rec(PPDB * db, void *rec, size_t expand)
{
	struct ppdb_value_variable *vv;
	int old, new;
	vv = (typeof(vv)) ((char *) rec - offsetof(typeof(*vv), value));
	old = vv->vlen + offsetof(typeof(*vv), value);
	new = old + expand;
	if (new > (1 << (vv->order + ppdb_variable_shift_min))) {
		struct ppdb_value_variable *vv_new = get_free_val(db, new);
		unsigned int *kvoff = ppdb_pointer(db, vv->kvoff);
		vv_new->vlen = vv->vlen;
		vv_new->kvoff = vv->kvoff;
		*kvoff = ppdb_offset(db, vv_new);
		memcpy(&(vv_new->value), &(vv->value), vv->vlen);
		put_free_val(db, vv);
		vv = vv_new;
	}
	memset((char *) &(vv->value) + vv->vlen, 0, expand);
	vv->vlen += expand;
	return (&(vv->value));
}

/* Truncate a record, always in place. */

static void
ppdb_truncate_rec(PPDB * db, void *rec, size_t newsize)
{
	struct ppdb_value_variable *vv;
	void *p = (char *) rec - offsetof(typeof(*vv), value);
	vv = (struct ppdb_value_variable *) p;
	vv->vlen = newsize;
}

/* Insert a new database record, both key and data are variable sized */

static void
new_db_rec(PPDB * db, void *key, unsigned int ksize, void *val,
	   unsigned int vsize)
{
	datum k, v;
	k.dptr = key;
	k.dsize = ksize;
	v.dptr = val;
	v.dsize = vsize;
	ppdb_store(db, k, v, PPDB_INSERT);
}

/*============================================================================*/

/* Add a text record to the database.  It creates a text record and sets the
 * token.  The token is set before the new record is added so the token pointer
 * is not invalidated before it is set, although it may be invalid after adding
 * the new record.
 */

DB_TOKEN
new_db_text(PPDB * db, const char *text, DB_TOKEN * token_p)
{
	datum value;
	DB_TOKEN token;
	assign_db_key(db, DB_RECTYPE_TEXT, token_p);
	value.dptr = (char *) text;
	value.dsize = strlen(text) + 1;
	token = *token_p;
	insert_db_rec_variable(db, token, value);
	return (token);
}

/* Read a text record from the database */

char *
read_db_text(PPDB * db, DB_TOKEN token)
{
	assert(token.type == DB_RECTYPE_TEXT);
	return (read_db_rec(db, token).dptr);
}

/* Update the text value for a text token, which must exist.  This may extend
 * the database and invalidate pointers.
 */

void
update_db_text(PPDB * db, const char *text, DB_TOKEN token)
{
	datum value;
	assert(token.type == DB_RECTYPE_TEXT);
	value.dptr = (char *) text;
	value.dsize = strlen(text) + 1;
	replace_db_rec_variable(db, token, value);
}

/* Empty a text record, which must exist. */

void
empty_db_text(PPDB * db, DB_TOKEN token)
{
	datum value;
	static const char empty_text[] = "";
	assert(token.type == DB_RECTYPE_TEXT);
	value.dptr = (char *) empty_text;
	value.dsize = 1;
	replace_db_rec_variable(db, token, value);
}

/*============================================================================*/

/* Scan a filename record looking for an entry of the required type with the
 * required index.
 */

static DB_REC_FILENAME_1 *
scan_db_filename(DB_REC_FILENAME_LIST * record, DB_TYPE type, int index)
{
	int i;
	DB_REC_FILENAME_1 *entry;
	for (i = 0, entry = record->list; i < record->count; ++i, ++entry) {
		if (entry->index == index && entry->entry.type == type) {
			return (entry);
		}
	}
	return (NULL);
}

/* Scan a filename record looking for an entry of the required type with the
 * highest index.
 */

static DB_REC_FILENAME_1 *
scan_db_filename_highest(DB_REC_FILENAME_LIST * record, DB_TYPE type)
{
	int i;
	DB_REC_FILENAME_1 *entry, *highest = NULL;
	for (i = 0, entry = record->list; i < record->count; ++i, ++entry) {
		if (entry->entry.type == type) {
			if (!highest || highest->index < entry->index) {
				highest = entry;
			}
		}
	}
	return (highest);
}

/* Add an entry to a filename record. */

static void
add_db_filename_list(PPDB * db, DB_REC_FILENAME_LIST ** record, int index,
		     DB_TOKEN token)
{
	DB_REC_FILENAME_1 new = {
		.index = index,
		.entry = token
	};
	assert(scan_db_filename(*record, token.type, index) == NULL);
	++((*record)->count);
	*record = ppdb_expand_rec(db, *record, sizeof (new));
	(*record)->list[(*record)->count - 1] = new;
}

/* Update filename data.  The record must already exist. */

static void
update_db_filename(PPDB * db, const char *filename,
		   const DB_REC_FILENAME_LIST * record)
{
	datum key, val;
	key.dptr = (char *) filename;
	key.dsize = strlen(filename) + 1;
	val.dptr = (void *) record;
	val.dsize = DB_SIZE_FILENAME_LIST(record);
	ppdb_store(db, key, val, PPDB_REPLACE);
}

/* Add an entry with the specified type and index to a filename record.
 * Add the record for the entry to the database, with zero contents except for
 * the relname pointer.
 * Requires that all records have DB_TOKEN relname at offset 0.
 */

static DB_TOKEN
new_db_filename_type(PPDB * db, const char *name, DB_TYPE type, int index,
		     size_t record_size)
{
	DB_REC_FILENAME_LIST *filename_rec;
	DB_REC_FILENAME_1 *slot;
	DB_TOKEN token;
	unsigned char record[record_size];
	datum value;
	int l = strlen(name);
	char name2[l + 1];
	assert(type == DB_RECTYPE_SOURCE || type == DB_RECTYPE_TARGET
	       || type == DB_RECTYPE_DIRECTORY);
	memcpy(name2, name, l + 1);
	strip_trailing(name2, l);
	assign_db_key(db, type, &token);
	filename_rec = read_db_filename_list(db, name2, 1);
	add_db_filename_list(db, &filename_rec, index, token);
	update_db_filename(db, name2, filename_rec);
	memset(record, 0, record_size);
	slot = scan_db_filename(filename_rec, DB_RECTYPE_TEXT, 0);
	/* First field of *record must be relname */
	assert(offsetof(DB_REC_SOURCE, relname) == 0);
	assert(offsetof(DB_REC_TARGET, relname) == 0);
	assert(offsetof(DB_REC_DIRECTORY, relname) == 0);
	/* Data should be aligned */
	*((DB_TOKEN *) record) = slot->entry;
	value.dptr = record;
	value.dsize = record_size;
	insert_db_rec_variable(db, token, value);
	return (token);
}

/*============================================================================*/

/* Update the child list for the parent directory of a source or target file or
 * for another directory.  Requires that source, target and directory records
 * have the same offsets for relname and parent.
 */

static void
update_db_parent_dir(PPDB * db, const char *name, int index, DB_TOKEN add)
{
	DB_REC_DIRECTORY *d;
	DB_REC_SOURCE *s;	/* Treat source/target/directory as source */
	DB_TOKEN s_parent, s_relname;
	DB_REC_CHILD_1 child;
	int l = strlen(name);
	char parent[l + 1];
	assert(offsetof(DB_REC_SOURCE, relname) ==
	       offsetof(DB_REC_TARGET, relname));
	assert(offsetof(DB_REC_SOURCE, relname) ==
	       offsetof(DB_REC_DIRECTORY, relname));
	assert(offsetof(DB_REC_SOURCE, parent) ==
	       offsetof(DB_REC_TARGET, parent));
	assert(offsetof(DB_REC_SOURCE, parent) ==
	       offsetof(DB_REC_DIRECTORY, parent));
	memcpy(parent, name, l + 1);
	l = strip_trailing(parent, l);
	if (l == 1) {
		return;		/* no parent for '/' */
	}
	while (l-- && parent[l] != '/') {
	};
	parent[l + 1] = '\0';
	s = (DB_REC_SOURCE *) (read_db_rec(db, add).dptr);
	if (DB_IS_NULL_TOKEN(s->parent)) {
		s->parent = lookup_db_directory(db, parent);
		if (DB_IS_NULL_TOKEN(s->parent)) {
			DB_TOKEN parent_p;
			parent_p = new_db_directory(db, parent);
			/* May have split, re-read source */
			s = (DB_REC_SOURCE *) (read_db_rec(db, add).dptr);
			s->parent = parent_p;
		}
	}
	s_parent = s->parent;
	s_relname = s->relname;
	d = read_db_directory(db, s_parent);
	if (DB_IS_NULL_TOKEN(d->children)) {
		new_db_child_list(db, &(d->children));
		d = read_db_directory(db, s_parent);	/* may have moved */
	}
	memset(&child, 0, sizeof (child));
	child.index = index;
	child.relname = s_relname;
	child.entry = add;
	add_db_child_list(db, d->children, &child, NULL);
	if (parent[1] && DB_IS_NULL_TOKEN(d->parent)) {
		/* Recursively update my non-root parent */
		update_db_parent_dir(db, parent, objtree_treename_index,
				     s_parent);
	}
}

/*============================================================================*/

/* Read a target record by key, key must exist */

DB_REC_TARGET *
read_db_target(PPDB * db, DB_TOKEN token)
{
	DB_REC_TARGET *t;
	datum val;
	assert(token.type == DB_RECTYPE_TARGET);
	val = read_db_rec(db, token);
	assert(val.dsize = sizeof (*t));
	t = (DB_REC_TARGET *) (val.dptr);
	return (t);
}

/* Create a new target record. */

DB_TOKEN
new_db_target(PPDB * db, const char *name)
{
	DB_TOKEN token;
	token =
	    new_db_filename_type(db, name, DB_RECTYPE_TARGET,
				 objtree_treename_index,
				 sizeof (DB_REC_TARGET));
	update_db_parent_dir(db, name, objtree_treename_index, token);
	return (token);
}

/* Find a target record in the database, create a new one if it does not exist.
 * All targets are also registered as sources, they can be included.
 */

DB_TOKEN
findnew_db_target(PPDB * db, const char *name)
{
	DB_TOKEN token = lookup_db_target(db, name);
	if (DB_IS_NULL_TOKEN(token)) {
		token = new_db_target(db, name);
	} else {
		update_db_parent_dir(db, name, objtree_treename_index, token);
	}
	findnew_db_source(db, name, objtree_treename_index);
	return (token);
}

/*============================================================================*/

/* The list processing routines are almost identical, apart from names, types
 * and the odd little bit of code.  Instantiate all list processing via one big
 * macro.  U is the type (upper case), l is the name (lower case).  Before
 * invoking this macro, define 
 *
 * __empty_db_##l##_list	Set flags after creating or emptying the list
 * __add_db_##l##_list_a	Assert type of entry allowed in list
 * __add_db_##l##_list_b	Set flags after adding an entry
 *
 * C++ is not the answer, if the answer is C++ then you are asking the wrong
 * question.
 */

#define instantiate_db_list(U, l) 						\
										\
/* Assign a new, empty ##U##_list record */					\
										\
void 										\
new_db_##l##_list(PPDB * db, DB_TOKEN *list_p) 					\
{ 										\
	DB_REC_##U##_LIST d, *c = &d;						\
	DB_TOKEN token;								\
	assert(DB_IS_NULL_TOKEN(*list_p)); 					\
	memset(c, 0, sizeof (*c)); 						\
	__empty_db_##l##_list 							\
	assign_db_key(db, DB_RECTYPE_##U##_LIST, list_p); 			\
	token = *list_p;							\
	new_db_rec(db, &token, sizeof(token), c, sizeof(*c)); 			\
} 										\
										\
/* Read data for a ##U##_list key (which must exist) from the database */	\
										\
DB_REC_##U##_LIST * 								\
read_db_##l##_list(PPDB * db, DB_TOKEN list) 					\
{ 										\
	datum value; 								\
	DB_REC_##U##_LIST *c;	 						\
	assert(list.type == DB_RECTYPE_##U##_LIST);				\
	value = read_db_rec(db, list);	 					\
	c = value.dptr;								\
	assert(value.dsize == DB_SIZE_##U##_LIST(c)); 				\
	return (c); 								\
} 										\
										\
/* Under most circumstances it is safe to read and cache a pointer to a list	\
 * entry, the cached entry is not affected by most changes to the database.	\
 * However it is not safe to cache a pointer over calls to code that can	\
 * expand entries of the same list type.  It is possible for the called code	\
 * to expand the cached entry itself which invalidates the pointer.  For such	\
 * code, take a copy of the record, which must then be treated as read only,	\
 * changes to the copy have no effect on the database.  The caller must free	\
 * the copy.									\
 */                                                                             \
										\
DB_REC_##U##_LIST * 								\
copy_db_##l##_list(PPDB * db, DB_TOKEN list)					\
{ 										\
	DB_REC_##U##_LIST *c, *copy;						\
	size_t size;								\
	assert(list.type == DB_RECTYPE_##U##_LIST);				\
	c = read_db_##l##_list(db, list);					\
	size = DB_SIZE_##U##_LIST(c);						\
	copy = k_malloc(size);							\
	memcpy(copy, c, size);							\
	return (copy);								\
} 										\
										\
/* Empty a ##U##_list, which must exist. */ 					\
										\
void 										\
empty_db_##l##_list(PPDB * db, DB_TOKEN list)					\
{ 										\
	DB_REC_##U##_LIST *c =							\
		read_db_##l##_list(db, list);			 		\
	c->count = 0; 								\
	__empty_db_##l##_list 							\
	ppdb_truncate_rec(db, c, offsetof(DB_REC_##U##_LIST, list));		\
} 										\
										\
/* Add a ##U##_1 to a ##U##_list, which must exist.  Duplicate entries are not	\
 * added, unless the entry has a null type (placeholder for later expansion).	\
 */						 				\
										\
static int added_db_##l##_list;							\
										\
DB_REC_##U##_LIST * 								\
add_db_##l##_list(PPDB * db, DB_TOKEN list,					\
		const DB_REC_##U##_1 *add, DB_REC_##U##_1 **added) 		\
{ 										\
	DB_REC_##U##_1 *c1 = NULL; 						\
	DB_REC_##U##_LIST *c = read_db_##l##_list(db, list);			\
	int i; 									\
	__add_db_##l##_list_a							\
	if (add->entry.type) {							\
		for (i = 0, c1 = c->list; i < c->count; ++i, ++c1) {		\
			if (memcmp(&(add->entry), &(c1->entry),			\
						sizeof (add->entry)) == 0) {	\
				break; 						\
			} 							\
		}								\
	} else {								\
		i = c->count;							\
	}									\
	if (i == c->count) { 							\
		c = ppdb_expand_rec(db, c, sizeof(*c1));			\
		added_db_##l##_list = 1;					\
		++(c->count); 							\
		__add_db_##l##_list_b 						\
		c1 = c->list + (c->count - 1); 					\
		*c1 = *add; 							\
	} 									\
	if (added) {								\
		*added = c1;							\
	}									\
	return (c); 								\
}

#define __empty_db_base_list
#define __add_db_base_list_a
#define __add_db_base_list_b
instantiate_db_list(BASE, base)
#define __empty_db_target_list
#define __add_db_target_list_a	assert(add->entry.type == DB_RECTYPE_SOURCE ||	\
				add->entry.type == DB_RECTYPE_TARGET);
#define __add_db_target_list_b
    instantiate_db_list(TARGET, target)
#define __empty_db_ccd_list	c->flags.sorted = 1;
#define __add_db_ccd_list_a	assert(add->entry.type == DB_RECTYPE_SOURCE);
#define __add_db_ccd_list_b	c->flags.sorted = 0;
    instantiate_db_list(CCD, ccd)
#define __empty_db_select_list
#define __add_db_select_list_a	assert(add->entry.type == DB_RECTYPE_TARGET ||	\
				add->entry.type == DB_RECTYPE_DIRECTORY);
#define __add_db_select_list_b
    instantiate_db_list(SELECT, select)
#define __empty_db_base_select_list
#define __add_db_base_select_list_a
#define __add_db_base_select_list_b
    instantiate_db_list(BASE_SELECT, base_select)
#define __empty_db_child_list	c->flags.sorted = 1;
#define __add_db_child_list_a	assert(add->entry.type == DB_RECTYPE_SOURCE ||	\
				add->entry.type == DB_RECTYPE_TARGET ||		\
				add->entry.type == DB_RECTYPE_DIRECTORY);
#define __add_db_child_list_b	c->flags.sorted = 0;
    instantiate_db_list(CHILD, child)

/*============================================================================*/
/* Scan the base record looking for a base name */
DB_REC_BASE_1 *
scan_db_base(PPDB * db, const char *name)
{
	int i;
	const char *base_name;
	for (i = 0; i < rec_base->count; ++i) {
		base_name = read_db_text(db, rec_base->list[i].entry);
		if (strcmp(base_name, name) == 0) {
			return (rec_base->list + i);
		}
	}
	return (NULL);
}

/* Resize all base_select records to reflect the base list.  Only values can
 * move, not the keys, so we only need one pass over the database.
 */

static int
resize_db_base_select_run(PPDB * db, kvpair kv, void *userdata)
{
	DB_TOKEN token;
	DB_REC_BASE_SELECT_LIST *bs;
	memcpy(&token, (void *) (kv.key.dptr), sizeof (token));
	if (token.type == DB_RECTYPE_BASE_SELECT_LIST) {
		DB_REC_BASE_SELECT_1 b1;
		memset(&b1, 0, sizeof (b1));
		bs = (DB_REC_BASE_SELECT_LIST *) (kv.val.dptr);
		assert(kv.val.dsize = DB_SIZE_BASE_SELECT_LIST(bs));
		while (bs->count < rec_base->count) {
			bs = add_db_base_select_list(db, token, &b1, NULL);
		}
	}
	return (0);
}

static void
resize_db_base_select(PPDB * db)
{
	run_db(db, resize_db_base_select_run, NULL);
}

/* Add a base to the base record. */

void
add_db_base(PPDB * db, const char *name, int has_modules, int has_arch_head,
	    int is_host, DB_TOKEN topdir)
{
	DB_REC_BASE_1 b1;
	memset(&b1, 0, sizeof (b1));
	b1.flags.has_modules = has_modules;
	b1.flags.has_arch_head = has_arch_head;
	b1.flags.is_host = is_host;
	new_db_text(db, name, &(b1.entry));
	b1.topdir = topdir;
	add_db_base_list(db, key_base, &b1, NULL);
	rec_base = read_db_base_list(db, key_base);	/* may have moved */
	resize_db_base_select(db);	/* adjust directory sizes to match base list */
}

/* Initialize the base record in the main database. */

static void
init_db_base(PPDB * db)
{
	DB_REC_BASE_LIST rb;
	memset(&rb, 0, sizeof (rb));
	insert_db_rec_fixed(db, key_base, rb);
}

/*============================================================================*/

/* Find a filename key in the database and return the data.  If the
 * key does not exist and 'create' is set then create the record.
 */

DB_REC_FILENAME_LIST *
read_db_filename_list(PPDB * db, const char *filename, int create)
{
	datum key, value;
	DB_REC_FILENAME_LIST *rec;
	key.dptr = (char *) filename;
	key.dsize = strlen(filename) + 1;
	value = ppdb_fetch(db, key);
	if (!value.dptr && !create) {
		return (NULL);
	}
	if (!value.dptr) {
		/* Start off with a DB_TOKEN to the text string */
		struct {
			DB_REC_FILENAME_LIST header;
			DB_REC_FILENAME_1 data;
		} new;
		new.header.count = 1;
		new_db_text(db, filename, &(new.data.entry));
		new.data.index = 0;
		value.dptr = &new;
		value.dsize = DB_SIZE_FILENAME_LIST(&(new.header));
		ppdb_store(db, key, value, PPDB_INSERT);
		value = ppdb_fetch(db, key);
		assert(value.dptr);
	}
	rec = value.dptr;
	assert(value.dsize == DB_SIZE_FILENAME_LIST(rec));
	return (rec);
}

/* Lookup a record by type, name and index */

DB_TOKEN
lookup_db_byname(PPDB * db, const char *name, DB_TYPE type, int index)
{
	DB_REC_FILENAME_LIST *filename_rec;
	DB_REC_FILENAME_1 *slot;
	DB_TOKEN token = db_null_token;
	int l = strlen(name);
	char name2[l + 1];
	memcpy(name2, name, l + 1);
	strip_trailing(name2, l);
	filename_rec = read_db_filename_list(db, name2, 0);
	if (filename_rec) {
		slot = scan_db_filename(filename_rec, type, index);
		if (slot) {
			token = slot->entry;
		}
	}
	return (token);
}

/* Lookup a record by type and name, return the entry with the highest index */

DB_TOKEN
lookup_db_byname_highest(PPDB * db, const char *name, DB_TYPE type)
{
	DB_REC_FILENAME_LIST *filename_rec;
	DB_REC_FILENAME_1 *slot;
	DB_TOKEN token = db_null_token;
	int l = strlen(name);
	char name2[l + 1];
	memcpy(name2, name, l + 1);
	strip_trailing(name2, l);
	filename_rec = read_db_filename_list(db, name2, 0);
	if (filename_rec) {
		slot = scan_db_filename_highest(filename_rec, type);
		if (slot) {
			token = slot->entry;
		}
	}
	return (token);
}

/*============================================================================*/

struct qsort_directory_t {
	DB_REC_CHILD_1 child;
	char *relname_text;
};

static int
qsort_db_directory(const void *va, const void *vb)
{
	const struct qsort_directory_t *a = va, *b = vb;
	int ret;
	if (a->child.entry.type == DB_RECTYPE_DIRECTORY
	    && b->child.entry.type != DB_RECTYPE_DIRECTORY) {
		return (1);	/* files first, then directories */
	}
	if (a->child.entry.type != DB_RECTYPE_DIRECTORY
	    && b->child.entry.type == DB_RECTYPE_DIRECTORY) {
		return (-1);
	}
	if ((ret = strcmp(a->relname_text, b->relname_text))) {
		return (ret);	/* by name */
	}
	if (a->child.index < b->child.index) {
		return (1);	/* descending index within name */
	}
	if (a->child.index > b->child.index) {
		return (-1);
	}
	if (a->child.entry.type == DB_RECTYPE_SOURCE
	    && b->child.entry.type != DB_RECTYPE_SOURCE) {
		return (-1);	/* source befoer target */
	}
	assert(a->child.entry.type == DB_RECTYPE_TARGET
	       && b->child.entry.type == DB_RECTYPE_SOURCE);
	return (1);
}

static int
sort_db_directories_run(PPDB * db, kvpair kv, void *userdata)
{
	DB_TOKEN token;
	DB_REC_DIRECTORY *d;
	DB_REC_CHILD_LIST *c;
	DB_REC_CHILD_1 *c1;
	memcpy(&token, (void *) (kv.key.dptr), sizeof (token));
	if (token.type != DB_RECTYPE_DIRECTORY) {
		return (0);
	}
	d = (DB_REC_DIRECTORY *) (kv.val.dptr);
	if (DB_IS_NULL_TOKEN(d->children)) {
		return (0);
	}
	c = read_db_child_list(db, d->children);
	if (!c->flags.sorted) {
		struct qsort_directory_t sort_data[c->count], *sd;
		int i;
		for (i = 0, sd = sort_data, c1 = c->list; i < c->count;
		     ++i, ++sd, ++c1) {
			sd->child = *c1;
			sd->relname_text =
			    k_strdup(read_db_text(db, c1->relname));
		}
		qsort(sort_data, c->count, sizeof (sort_data[0]),
		      qsort_db_directory);
		c->flags.sorted = 1;
		for (i = 0, sd = sort_data, c1 = c->list; i < c->count;
		     ++i, ++sd, ++c1) {
			*c1 = sd->child;
			c1->flags.primary = (i == 0
					     || memcmp(&((c1 - 1)->relname),
						       &(c1->relname),
						       sizeof (c1->relname)));
			free(sd->relname_text);
		}
	}
	return (0);
}

int
sort_db_directories(PPDB * db)
{
	return (run_db(db, sort_db_directories_run, NULL));
}

/* Read a directory record by key, key must exist */

DB_REC_DIRECTORY *
read_db_directory(PPDB * db, DB_TOKEN token)
{
	DB_REC_DIRECTORY *d;
	datum val;
	assert(token.type == DB_RECTYPE_DIRECTORY);
	val = read_db_rec(db, token);
	assert(val.dsize = sizeof (*d));
	d = (DB_REC_DIRECTORY *) (val.dptr);
	return (d);
}

/* Create a directory record, always assign a base_select list */

DB_TOKEN
new_db_directory(PPDB * db, const char *name)
{
	DB_TOKEN token;
	DB_REC_DIRECTORY *d = NULL;
	DB_REC_BASE_SELECT_LIST *bs, temp;
	datum val;
	unsigned int size;
	token = new_db_filename_type(db, name, DB_RECTYPE_DIRECTORY,
				     objtree_treename_index,
				     DB_SIZE_DIRECTORY(d));
	update_db_parent_dir(db, name, objtree_treename_index, token);
	val = read_db_rec(db, token);
	d = val.dptr;
	/* Size the base_select list according to the current number of bases */
	temp.count = rec_base ? rec_base->count : 0;
	size = DB_SIZE_BASE_SELECT_LIST(&temp);
	bs = k_malloc(size);
	memset(bs, 0, size);
	bs->count = temp.count;
	assign_db_key(db, DB_RECTYPE_BASE_SELECT_LIST, &(d->base_select));
	new_db_rec(db, &(d->base_select), sizeof (d->base_select), bs, size);
	free(bs);
	return (token);
}

/* Find a directory record in the database, create a new one if it does not exist. */

DB_TOKEN
findnew_db_directory(PPDB * db, const char *name)
{
	DB_TOKEN token;
	token =
	    lookup_db_byname(db, name, DB_RECTYPE_DIRECTORY,
			     objtree_treename_index);
	if (DB_IS_NULL_TOKEN(token)) {
		token = new_db_directory(db, name);
	} else {
		update_db_parent_dir(db, name, objtree_treename_index, token);
	}
	return (token);
}

/*============================================================================*/

/* Read a source record by key, key must exist */

DB_REC_SOURCE *
read_db_source(PPDB * db, DB_TOKEN token)
{
	DB_REC_SOURCE *s;
	datum val;
	assert(token.type == DB_RECTYPE_SOURCE);
	val = read_db_rec(db, token);
	assert(val.dsize = sizeof (*s));
	s = (DB_REC_SOURCE *) (val.dptr);
	return (s);
}

/* Create a source record. */

DB_TOKEN
new_db_source(PPDB * db, const char *name, int index)
{
	DB_TOKEN token;
	token = new_db_filename_type(db, name, DB_RECTYPE_SOURCE, index,
				     sizeof (DB_REC_SOURCE));
	update_db_parent_dir(db, name, index, token);
	read_db_source(db, token)->index = index;
	return (token);
}

/* Find a source record in the database, create a new one if it does not exist. */

DB_TOKEN
findnew_db_source(PPDB * db, const char *name, int index)
{
	DB_TOKEN token;
	token = lookup_db_byname(db, name, DB_RECTYPE_SOURCE, index);
	if (DB_IS_NULL_TOKEN(token)) {
		token = new_db_source(db, name, index);
	} else {
		update_db_parent_dir(db, name, index, token);
	}
	return (token);
}

/*============================================================================*/

/* Run the entire database, calling the user function.   It returns all the
 * fixed keys first then the variable keys.  If the user function adds new fixed
 * keys, they will only be seen by the user function if the code has not gone
 * onto the variable length keys.
 */

int
run_db(PPDB * db, int (*userfunc) (PPDB *, kvpair, void *), void *userdata)
{
	kvpair kv;
	int ret = 0;
	const static kvpair nullkv;
	if (signalled) {
		return (0);
	}
	for (kv = ppdb_first(db); memcmp(&kv, &nullkv, sizeof (kv));
	     kv = ppdb_next(db)) {
		if ((ret = userfunc(db, kv, userdata))) {
			break;
		}
	}
	return (ret);
}

/* Run the entire database in sorted directory order using the child lists,
 * calling the user function.  If the user function changes the size of a 
 * child list while it is being processed then this code resets to the start
 * of the particular list.
 *
 * Changes to the child list for the current directory, sub directories or
 * parent directories will be processed by this routine.  Changes to unrelated
 * directories may or may not be processed by this routine.
 */

int run_db_child_depth;

static int
run_db_child_recurse(PPDB * db,
		     int (*userfunc) (PPDB *, DB_REC_CHILD_1 *, void *),
		     void *userdata, const char *relname)
{
	DB_TOKEN token;
	DB_REC_DIRECTORY *d;
	DB_REC_CHILD_LIST *c;
	DB_REC_CHILD_1 *c1;
	int i, ret = 0, save_count;
	const char *dir_relname;
	if (signalled) {
		return (0);
	}
	token = lookup_db_directory(db, relname);
	d = read_db_directory(db, token);
	if (DB_IS_NULL_TOKEN(d->children)) {
		return (0);
	}
	++run_db_child_depth;
	assert(run_db_child_depth < 1000);	/* arbitrary loop detection */
	c = read_db_child_list(db, d->children);
	assert(c->flags.sorted);
	do {
		added_db_child_list = 0;
		save_count = c->count;
		for (i = 0; i < save_count; ++i) {
			if (added_db_child_list) {
				added_db_child_list = 0;
				c = read_db_child_list(db, d->children);
				assert(c->flags.sorted);
				if (save_count != c->count) {
					break;	/* redo from start */
				}
			}
			c1 = c->list + i;
			ret = userfunc(db, c1, userdata);
			if (!ret) {
				if (c1->entry.type != DB_RECTYPE_DIRECTORY) {
					continue;
				}
				dir_relname = read_db_text(db, c1->relname);
				ret =
				    run_db_child_recurse(db, userfunc, userdata,
							 dir_relname);
			}
			if (ret == 2) {
				ret = 0;	/* pruned */
			}
		}
	} while (save_count != c->count);
	--run_db_child_depth;
	return (ret);
}

int
run_db_child(PPDB * db, int (*userfunc) (PPDB *, DB_REC_CHILD_1 *, void *),
	     void *userdata, const char *relname)
{
	run_db_child_depth = 0;
	return (run_db_child_recurse(db, userfunc, userdata, relname));
}

/* Run the entire database in base select order using the base_select lists,
 * calling the user function.  If the user function changes the size of a 
 * base select list while it is being processed then this code resets to the
 * start of the particular list.
 *
 * Changes to the base select list for the current directory, sub directories or
 * parent directories will be processed by this routine.  Changes to unrelated
 * directories may or may not be processed by this routine.
 */

int run_db_select_depth;

static int
run_db_select_recurse(PPDB * db,
		      int (*userfunc) (PPDB *, DB_REC_BASE_1 *,
				       DB_REC_SELECT_1 *, void *),
		      void *userdata, DB_REC_BASE_1 * base1,
		      DB_REC_DIRECTORY * d, int base_number)
{
	DB_TOKEN s_token;
	DB_REC_SELECT_LIST *s;
	DB_REC_SELECT_1 *s1;
	DB_REC_BASE_SELECT_LIST *bs;
	int i, ret = 0, save_count;
	if (signalled) {
		return (0);
	}
	bs = read_db_base_select_list(db, d->base_select);
	s_token = bs->list[base_number].entry;
	if (DB_IS_NULL_TOKEN(s_token)) {
		return (0);
	}
	++run_db_select_depth;
	assert(run_db_select_depth < 1000);	/* arbitrary loop detection */
	s = read_db_select_list(db, s_token);
	do {
		added_db_select_list = 0;
		save_count = s->count;
		for (i = 0; i < save_count; ++i) {
			if (added_db_select_list) {
				added_db_select_list = 0;
				s = read_db_select_list(db, s_token);
				if (save_count != s->count) {
					break;	/* redo from start */
				}
			}
			s1 = s->list + i;
			ret = userfunc(db, base1, s1, userdata);
			if (!ret) {
				if (s1->entry.type != DB_RECTYPE_DIRECTORY
				    || !s1->flags.sels) {
					continue;
				}
				ret =
				    run_db_select_recurse(db, userfunc,
							  userdata, base1,
							  read_db_directory(db,
									    s1->
									    entry),
							  base_number);
			}
			if (ret == 2) {
				ret = 0;	/* pruned */
			}
		}
	} while (save_count != s->count);
	--run_db_select_depth;
	return (ret);
}

int
run_db_select(PPDB * db,
	      int (*userfunc) (PPDB *, DB_REC_BASE_1 *, DB_REC_SELECT_1 *,
			       void *), void *userdata, DB_REC_BASE_1 * base1)
{
	DB_REC_DIRECTORY *d;
	int base_number = base1 - rec_base->list;
	assert(base_number >= 0 && base_number < rec_base->count);
	d = read_db_directory(db, base1->topdir);
	run_db_select_depth = 0;
	return (run_db_select_recurse(db, userfunc, userdata, base1, d,
				      base_number));
}

/*============================================================================*/

/* Lock the main pre-processing database.  To ensure that it is locked even over
 * NFS, create a temporary file based on the current pid, link the lock file to
 * the temporary file then check that both files have the same stat data.
 *
 * Requires compile variables PP_LOCK1 (lock filename), PP_LOCK2 (prefix
 * of temporary filename), PP_LOCK_SLEEP (how to sleep when waiting for
 * the lock).
 */

static int have_lock;

static void
lock_db_main(void)
{
	struct stat s1, s2;
	int fd, pid, prevpid = 0, sleep = 0;
	FILE *f;
	char name[PATH_MAX + 1];
	snprintf(name, sizeof (name), PP_LOCK2 ".%d", getpid());
	unlink(name);
	if ((fd = open(name, O_CREAT | O_EXCL | O_WRONLY, 0666)) < 0) {
		fprintf(stderr, "%s: Creation of %s failed: %m\n", program,
			name);
		exit(1);	/* no core dump */
	}
	have_lock = 1;
	fchmod(fd, 0666);
	f = fdopen(fd, "w");
	fprintf(f, "%ld\n", (unsigned long) getpid());
	fclose(f);
	while (!signalled) {
		if (sleep) {
			PP_LOCK_SLEEP;
		}
		sleep = 1;
		if (link(name, PP_LOCK1) && errno != EEXIST) {
			fprintf(stderr, "%s: link(%s, %s) failed: %m\n",
				program, name, PP_LOCK1);
			exit(1);	/* no core dump */
		}
		if (stat(name, &s2)) {
			fprintf(stderr, "%s: stat(%s) failed: %m\n", program,
				name);
			exit(1);	/* no core dump */
		}
		memset(&s1, 0, sizeof (s1));
		if (stat(PP_LOCK1, &s1) && errno != ENOENT) {
			fprintf(stderr, "%s: stat(%s) failed: %d %m\n", program,
				PP_LOCK1, errno);
			exit(1);	/* no core dump */
		}
		if (s1.st_dev == s2.st_dev && s1.st_ino == s2.st_ino
		    && s1.st_nlink == 2 && s2.st_nlink == 2) {
			have_lock = 2;
			break;
		}
		/* Locked, check if the lock process is still running */
		f = fopen(PP_LOCK1, "r");
		if (!f) {
			/* no lock file, race with unlock, retry */
			continue;
		}
		fscanf(f, "%d", &pid);
		fclose(f);
		if (kill(pid, 0) == 0) {
			/* lock process is still running, retry */
			continue;
		}
		if (errno != ESRCH) {
			fprintf(stderr,
				"%s: Unexpected error from kill(%d, 0) on lock process: %m\n",
				program, pid);
			exit(1);	/* no core dump */
		}
		if (pid != prevpid) {
			/* change of lock pid, race with unlock, retry */
			prevpid = pid;
			continue;
		}
		fprintf(stderr, "%s: Could not get the database lock\n",
			program);
		exit(1);	/* no core dump */
	}
}

static void
unlock_db_main(void)
{
	char name[PATH_MAX + 1];
	if (have_lock == 2) {
		if (unlink(PP_LOCK1)) {
			fprintf(stderr, "%s: unlink(%s) failed: %m\n", program,
				PP_LOCK1);
			abort();
		}
		--have_lock;
	}
	if (have_lock == 1) {
		snprintf(name, sizeof (name), PP_LOCK2 ".%d", getpid());
		if (unlink(name)) {
			fprintf(stderr, "%s: unlink(%s) failed: %m\n", program,
				name);
			abort();
		}
		--have_lock;
	}
}

/* Close the database and release the lock */

void
close_db_main(PPDB * db)
{
	if (db->header) {
		db->header->flags.dirty = 0;
	}
	rec_base = NULL;
	ppdb_close(db);
	if (close(db->fd)) {
		fprintf(stderr, "%s: close_db_main close failed: %m\n",
			program);
		abort();
	}
	unlock_db_main();
	free(db);
}

/* Initialize the main database */

static struct ppdb_header current_header = {
	.version = DB_VERSION,
	.fixed_slab_num = ppdb_slab_num,
	.fixed_key_vector_size = ppdb_fixed_key_vector_size,
	.variable_shift_min = ppdb_variable_shift_min,
	.variable_shift_max = ppdb_variable_shift_max,
	.hashsize = ppdb_hashsize,
};

static void
init_db_main(PPDB * db)
{
	DB_TOKEN root;
	struct init_data {
		struct ppdb_header h;
		struct ppdb_key_fixed k[ppdb_slab_num];
		DB_REC_SOURCE s[ppdb_slab_num];
		DB_REC_TARGET t[ppdb_slab_num];
		DB_REC_DIRECTORY d[ppdb_slab_num];
		char f[ppdb_extend];	/* Arbitrary initial free space */
	};
	struct init_data *init;

	init = k_malloc(sizeof (*init));
	memset(init, 0, sizeof (init));
	memcpy(&(init->h), &current_header, sizeof (current_header));
	init->h.flags.dirty = 1;
	init->h.next_recno = key_base.recno + 1;
	init->h.self = &(init->h);
	init->h.key_fixed[0] = init->k;
	init->h.source_slab = init->s;
	init->h.source_slab_free = ppdb_slab_num;
	init->h.target_slab = init->t;
	init->h.target_slab_free = ppdb_slab_num;
	init->h.directory_slab = init->d;
	init->h.directory_slab_free = ppdb_slab_num;
	init->h.free_start = init->f;
	init->h.free_left = sizeof (init->f);

	safe_write(db->fd, 0, init, sizeof (*init), "init_db_main");
	free(init);

	ppdb_open(db);
	init_db_base(db);
	root = new_db_directory(db, "/");
	add_db_base(db, "vmlinux", 1, 1, 0, root);
	add_db_base(db, "host", 0, 0, 1, root);
	db->header->flags.dirty = 0;
	if (msync(db->header, db->header->size, MS_SYNC)) {
		fprintf(stderr, "%s: init_db_main msync failed: %m\n", program);
		abort();
	}
	ppdb_close(db);
}

/* Copy an entire database record.  From and to must be different tokens. */

extern void
copy_db_data(PPDB * db, DB_TOKEN to, DB_TOKEN from)
{
	datum val;
	assert(to.type == from.type);
	assert(to.recno != from.recno);
	val = read_db_rec(db, from);
	replace_db_rec_variable(db, to, val);
}

/* Open the main pre-processing database.  Create and initialize if it does not
 * exist or is back level or dirty.
 *
 * open_type: 0 - Create if it does not exist or is dirty.
 *            1 - Use an exising database, which must be clean.
 *            2 - Open for dumping only.
 */

PPDB *
open_db_main(int open_type)
{
	int loop = 0;
	PPDB *db;
	struct ppdb_header h;
	db = k_malloc(sizeof (*db));
	memset(db, 0, sizeof (*db));
	pagesizem1 = getpagesize() - 1;
	if (open_type != 2) {
		lock_db_main();
	}
	while (!signalled && !db->header) {
		if (loop && open_type != 0) {
			fprintf(stderr,
				"%s: %s must be valid but is not, giving up\n",
				program, ppdb_name);
			abort();
		}
		if (loop) {
			unlink(ppdb_name);
		}
		if (loop > 1) {
			fprintf(stderr,
				"%s: open_db_main is looping, giving up\n",
				program);
			break;
		}
		if ((db->fd =
		     open(ppdb_name, O_RDWR | (open_type ? 0 : O_CREAT),
			  0666)) < 0) {
			fprintf(stderr, "%s: Cannot open %s for update: %m\n",
				program, ppdb_name);
			break;
		}
		if (loop) {
			init_db_main(db);
		}
		if (safe_read(db->fd, 0, &h, sizeof (h), "open_db_main")) {
			struct stat st;
			if (fstat(db->fd, &st)) {
				fprintf(stderr, "%s: fstat(%s) failed: %m\n",
					program, ppdb_name);
				abort();
			}
			if (st.st_size) {
				fprintf(stderr,
					"%s: %s has missing header, forcing complete rebuild\n",
					program, ppdb_name);
			}
		} else
		    if (memcmp
			(&h, &current_header,
			 offsetof(typeof(current_header), self))) {
			fprintf(stderr,
				"%s: %s critical header fields have changed, forcing complete rebuild\n",
				program, ppdb_name);
		} else if (h.flags.dirty && open_type != 2) {
			fprintf(stderr,
				"%s: %s was not shut down correctly, forcing complete rebuild\n",
				program, ppdb_name);
		} else {
			ppdb_open(db);
			if (open_type != 2) {
				db->header->flags.dirty = 1;
			}
			rec_base = read_db_base_list(db, key_base);
		}
		++loop;
	}
	if (!db->header) {
		++errors;
		if (open_type != 2) {
			unlock_db_main();
		}
		free(db);
		return (NULL);
	}
	return (db);
}
